from collections import Counter
from enum import Enum
from typing import TYPE_CHECKING, Callable, List, Union

import pandas as pd

from vivarium.framework.event import Event
from vivarium.framework.results.context import ResultsContext

if TYPE_CHECKING:
    # Cyclic import
    from vivarium.framework.engine import Builder


class SourceType(Enum):
    COLUMN = 0
    VALUE = 1


class ResultsManager:
    """Backend manager object for the results management system.

    The :class:`ResultManager` actually performs the actions needed to
    stratify and observe results. It contains the public methods used by the
    :class:`ResultsInterface` to register stratifications and observations,
    which provide it with lists of methods to apply in their respective areas.
    It is able to record observations at any of the time-step sub-steps
    (`time_step__prepare`, `time_step`, `time_step__cleanup`, and
    `collect_metrics`).
    """

    configuration_defaults = {
        "stratification": {
            "default": [],
        }
    }

    def __init__(self):
        self._metrics = Counter()
        self._results_context = ResultsContext()
        self._required_columns = {"tracked"}
        self._required_values = set()
        self._name = "results_manager"

    @property
    def name(self) -> str:
        """The name of this ResultsManager."""
        return self._name

    @property
    def metrics(self):
        return self._metrics.copy()

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: "Builder"):
        self.logger = builder.logging.get_logger(self.name)
        self.population_view = builder.population.get_view([])
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()

        builder.event.register_listener("time_step__prepare", self.on_time_step_prepare)
        builder.event.register_listener("time_step", self.on_time_step)
        builder.event.register_listener("time_step__cleanup", self.on_time_step_cleanup)
        builder.event.register_listener("collect_metrics", self.on_collect_metrics)

        self.get_value = builder.value.get_value

        self.set_default_stratifications(builder)

        builder.value.register_value_modifier("metrics", self.get_results)

    def on_time_step_prepare(self, event: Event):
        self.gather_results("time_step__prepare", event)

    def on_time_step(self, event: Event):
        self.gather_results("time_step", event)

    def on_time_step_cleanup(self, event: Event):
        self.gather_results("time_step__cleanup", event)

    def on_collect_metrics(self, event: Event):
        self.gather_results("collect_metrics", event)

    def gather_results(self, event_name: str, event: Event):
        population = self._prepare_population(event)
        for results_group in self._results_context.gather_results(population, event_name):
            self._metrics.update(results_group)

    def set_default_stratifications(self, builder):
        default_stratifications = builder.configuration.stratification.default
        self._results_context.set_default_stratifications(default_stratifications)

    def register_stratification(
        self,
        name: str,
        categories: List[str],
        mapper: Callable,
        is_vectorized: bool,
        requires_columns: List[str] = (),
        requires_values: List[str] = (),
    ) -> None:
        """Manager-level stratification registration, including resources and the stratification itself.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        categories
            List of string values that the mapper is allowed to output.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline modifier is called.

        Returns
        ------
        None
        """
        target_columns = list(requires_columns) + list(requires_values)
        self._results_context.add_stratification(
            name, target_columns, categories, mapper, is_vectorized
        )
        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

    def register_binned_stratification(
        self,
        target: str,
        target_type: str,
        binned_column: str,
        bins: List[Union[int, float]],
        labels: List[str],
        **cut_kwargs,
    ) -> None:
        """Manager-level registration of a continuous `target` quantity to observe into bins in a `binned_column`.

        Parameters
        ----------
        target
            String name of the state table column or value pipeline used to stratify.
        target_type
            "column" or "value"
        binned_column
            String name of the column for the binned quantities.
        bins
            List of scalars defining the bin edges, passed to :meth: pandas.cut. Lists
            `bins` and `labels` must be of equal length.
        labels
            List of string labels for bins. Lists `bins` and `labels` must be of equal length.
        **cut_kwargs
            Keyword arguments for :meth: pandas.cut.

        Returns
        ------
        None
        """

        def _bin_data(data: pd.Series) -> pd.Series:
            return pd.cut(data, bins, labels=labels, **cut_kwargs)

        if len(bins) != len(labels):
            raise ValueError(
                f"Bin length ({len(bins)}) does not match labels length ({len(labels)})"
            )
        target_arg = "required_columns" if target_type == "column" else "required_values"
        target_kwargs = {target_arg: [target]}
        self.register_stratification(
            binned_column, bins, _bin_data, is_vectorized=True, **target_kwargs
        )

    def register_observation(
        self,
        name: str,
        pop_filter: str,
        aggregator_sources: List[str],
        aggregator: Callable,
        requires_columns: List[str] = (),
        requires_values: List[str] = (),
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
        when: str = "collect_metrics",
    ) -> None:
        self._warn_check_stratifications(additional_stratifications, excluded_stratifications)
        self._results_context.add_observation(
            name,
            pop_filter,
            aggregator_sources,
            aggregator,
            additional_stratifications,
            excluded_stratifications,
            when,
        )
        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

    def _add_resources(self, target: List[str], target_type: SourceType):
        if not len(target):
            return  # do nothing on empty lists
        target = set(target) - {"event_time", "current_time", "step_size"}
        if target_type == SourceType.COLUMN:
            self._required_columns.update(target)
        elif target_type == SourceType.VALUE:
            self._required_values.update([self.get_value(t) for t in target])

    def _prepare_population(self, event: Event):
        population = self.population_view.subview(list(self._required_columns)).get(
            event.index
        )
        population["current_time"] = self.clock()
        population["step_size"] = event.step_size
        population["event_time"] = self.clock() + event.step_size
        for k, v in event.user_data.items():
            population[k] = v
        for pipeline in self._required_values:
            population[pipeline.name] = pipeline(event.index)
        return population

    def get_results(self, index, metrics):
        # Shim for now to allow incremental transition to new results system.
        metrics.update(self.metrics)
        return metrics

    def _warn_check_stratifications(
        self, additional_stratifications, excluded_stratifications
    ):
        """Check additional and excluded stratifications if they'd not affect
        stratifications (i.e., would be NOP), and emit warning."""
        nop_additional = [
            s
            for s in additional_stratifications
            if s in self._results_context.default_stratifications
        ]
        if len(nop_additional):
            self.logger.warning(
                f"Specified additional stratifications are already included by default: {nop_additional}",
            )
        nop_exclude = [
            s
            for s in excluded_stratifications
            if s not in self._results_context.default_stratifications
        ]
        if len(nop_exclude):
            self.logger.warning(
                f"Specified excluded stratifications are already not included by default: {nop_exclude}",
            )



"""
=====================
The Logging Subsystem
=====================

"""
from loguru import logger
from loguru._logger import Logger

from vivarium.framework.logging.utilities import configure_logging_to_terminal


class LoggingManager:
    def __init__(self):
        self._simulation_name = None

    def configure_logging(
        self,
        simulation_name: str,
        verbosity: int = 0,
        long_format: bool = True,
    ) -> None:
        self._simulation_name = simulation_name
        if self._terminal_logging_not_configured():
            configure_logging_to_terminal(verbosity=verbosity, long_format=long_format)

    @staticmethod
    def _terminal_logging_not_configured() -> bool:
        # This hacks into the internals of loguru to see if we've already configured a
        # terminal sink. Loguru maintains a global increment of the loggers it has generated
        # and has a default logger configured with id 0. All code paths in this library that
        # configure logging handlers delete the default handler with id 0, add a terminal
        # logging handler (with id 1) and potentially have a file logging handler with id 2.
        # This behavior is based on sequencing of the handle definition. This is a bit
        # fragile since it depends on a loguru's internals as well as the stability of code
        # paths in vivarium, but both are quite stable at this point, so I think it's pretty,
        # low risk.
        return 1 not in logger._core.handlers

    @property
    def name(self):
        return "logging_manager"

    def get_logger(self, component_name: str = None) -> Logger:
        bind_args = {"simulation": self._simulation_name}
        if component_name:
            bind_args["component"] = component_name
        return logger.bind(**bind_args)


class LoggingInterface:
    def __init__(self, manager: LoggingManager):
        self._manager = manager

    def get_logger(self, component_name: str = None) -> Logger:
        return self._manager.get_logger(component_name)



from typing import Dict

import pandas as pd

from vivarium.framework.engine import Builder


class Observer:
    configuration_defaults = {
        "mortality": {
            "life_expectancy": 80,
        }
    }

    def __init__(self):
        self.name = "observer"

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder):
        self.life_expectancy = builder.configuration.mortality.life_expectancy
        self.population_view = builder.population.get_view(["age", "alive"])

        builder.value.register_value_modifier("metrics", self.metrics)

    def metrics(self, index: pd.Index, metrics: Dict):
        pop = self.population_view.get(index)
        metrics["total_population_alive"] = len(pop[pop.alive == "alive"])
        metrics["total_population_dead"] = len(pop[pop.alive == "dead"])

        metrics["years_of_life_lost"] = (
            self.life_expectancy - pop.age[pop.alive == "dead"]
        ).sum()

        return metrics



import pandas as pd


class Risk:
    configuration_defaults = {
        "risk": {
            "proportion_exposed": 0.3,
        },
    }

    def __init__(self, name):
        self.name = name
        self.configuration_defaults = {name: Risk.configuration_defaults["risk"]}

    # noinspection PyAttributeOutsideInit
    def setup(self, builder):
        proportion_exposed = builder.configuration[self.name].proportion_exposed
        self.base_exposure_threshold = builder.value.register_value_producer(
            f"{self.name}.base_proportion_exposed",
            source=lambda index: pd.Series(proportion_exposed, index=index),
        )
        self.exposure_threshold = builder.value.register_value_producer(
            f"{self.name}.proportion_exposed", source=self.base_exposure_threshold
        )

        self.exposure = builder.value.register_value_producer(
            f"{self.name}.exposure", source=self._exposure
        )
        self.randomness = builder.randomness.get_stream(self.name)

        columns_created = [f"{self.name}_propensity"]
        builder.population.initializes_simulants(
            self.on_initialize_simulants,
            creates_columns=columns_created,
            requires_streams=[self.name],
        )
        self.population_view = builder.population.get_view(columns_created)

    def on_initialize_simulants(self, pop_data):
        draw = self.randomness.get_draw(pop_data.index)
        self.population_view.update(pd.Series(draw, name=f"{self.name}_propensity"))

    def _exposure(self, index):
        propensity = self.population_view.get(index)[f"{self.name}_propensity"]
        return self.exposure_threshold(index) > propensity

    def __repr__(self):
        return f"Risk(name={self.name})"


class RiskEffect:
    configuration_defaults = {
        "risk_effect": {
            "relative_risk": 2,
        },
    }

    def __init__(self, risk, disease_rate):
        self.risk = risk
        self.disease_rate = disease_rate
        self.name = f"effect_of_{risk}_on_{disease_rate}"
        self.configuration_defaults = {
            self.name: RiskEffect.configuration_defaults["risk_effect"]
        }

    # noinspection PyAttributeOutsideInit
    def setup(self, builder):
        relative_risk = builder.configuration[self.name].relative_risk
        self.relative_risk = builder.value.register_value_producer(
            f"{self.name}.relative_risk",
            source=lambda index: pd.Series(relative_risk, index=index),
        )

        builder.value.register_value_modifier(
            f"{self.disease_rate}.population_attributable_fraction",
            self.population_attributable_fraction,
        )
        builder.value.register_value_modifier(f"{self.disease_rate}", self.rate_adjustment)
        self.base_risk_exposure = builder.value.get_value(
            f"{self.risk}.base_proportion_exposed"
        )
        self.actual_risk_exposure = builder.value.get_value(f"{self.risk}.exposure")

    def population_attributable_fraction(self, index):
        exposure = self.base_risk_exposure(index)
        relative_risk = self.relative_risk(index)
        return exposure * (relative_risk - 1) / (exposure * (relative_risk - 1) + 1)

    def rate_adjustment(self, index, rates):
        exposed = self.actual_risk_exposure(index)
        rr = self.relative_risk(index)
        rates[exposed] *= rr[exposed]
        return rates



"""
================================
Population Management Exceptions
================================

Errors related to the mishandling of the simulation state table.
"""

from vivarium.exceptions import VivariumError


class PopulationError(VivariumError):
    """Error raised when the population is invalidly queried or updated."""

    pass



import sys

min_version, max_version = ((3, 8), "3.8"), ((3, 11), "3.11")

if not (min_version[0] <= sys.version_info[:2] <= max_version[0]):
    # Python 3.5 does not support f-strings
    py_version = ".".join([str(v) for v in sys.version_info[:3]])
    error = (
        "\n----------------------------------------\n"
        "Error: Vivarium runs under python {min_version}-{max_version}.\n"
        "You are running python {py_version}".format(
            min_version=min_version[1], max_version=max_version[1], py_version=py_version
        )
    )
    print(error, file=sys.stderr)
    sys.exit(1)

from pathlib import Path

from setuptools import find_packages, setup

if __name__ == "__main__":
    base_dir = Path(__file__).parent
    src_dir = base_dir / "src"

    about = {}
    with (src_dir / "vivarium" / "__about__.py").open() as f:
        exec(f.read(), about)

    with (base_dir / "README.rst").open() as f:
        long_description = f.read()

    install_requirements = [
        "numpy",
        "pandas<2.0.0",
        "pyyaml>=5.1",
        "scipy",
        "click",
        "tables",
        "networkx",
        "loguru",
    ]

    interactive_requirements = [
        "IPython",
        "ipywidgets",
        "jupyter",
    ]

    test_requirements = [
        "pytest",
        "pytest-mock",
    ]

    doc_requirements = [
        "sphinx>=4.0",
        "sphinx-rtd-theme",
        "sphinx-click",
        "IPython",
        "matplotlib",
    ]

    setup(
        name=about["__title__"],
        version=about["__version__"],
        description=about["__summary__"],
        long_description=long_description,
        license=about["__license__"],
        url=about["__uri__"],
        author=about["__author__"],
        author_email=about["__email__"],
        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: BSD",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            "Programming Language :: Python",
            "Programming Language :: Python :: Implementation :: CPython",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Life",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Software Development :: Libraries",
        ],
        package_dir={"": "src"},
        packages=find_packages(where="src"),
        include_package_data=True,
        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            "docs": doc_requirements,
            "test": test_requirements,
            "interactive": interactive_requirements,
            "dev": doc_requirements + test_requirements + interactive_requirements,
        },
        entry_points="""
                [console_scripts]
                simulate=vivarium.interface.cli:simulate
            """,
        zip_safe=False,
    )



__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "vivarium"
__summary__ = "vivarium is a microsimulation framework built on top of the standard scientific python stack."
__uri__ = "https://github.com/ihmeuw/vivarium"

__version__ = "1.2.0"

__author__ = "The vivarium developers"
__email__ = "vivarium.dev@gmail.com"

__license__ = "BSD-3-Clause"
__copyright__ = f"Copyright 2016-2023 {__author__}"
